"""Teacher Heper CLI"""

import code
import argparse

from .helper import Helper


helper = Helper.read_cache() if Helper.cache_exists() else None


def find_student(name):
    assert helper is not None
    print(helper.find_nearest_match(name, threshold=60))


def find_parent(name):
    assert helper is not None
    print(helper.find_parent(name))


def shell():
    code.interact(local={
        'helper': helper,
    })


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--student',
        '-s',
        help="Lookup a student and print the result"
    )
    parser.add_argument(
        '--parent',
        '-p',
        help="Lookup a parent and print the result"
    )
    parser.add_argument(
        '--new',
        action='store_const',
        const=True,
        help=('Regenerate the database by parsing student.csv and parent.csv '
              'in the $HELPER_DATA directory.')
    )

    args = parser.parse_args()

    if args.new:
        Helper.new_school_year().write_cache()
    elif args.student:
        find_student(args.student)
    elif args.parent:
        find_parent(args.parent)
    else:
        shell()


if __name__ == '__main__':
    main()



import csv
import unittest
from unittest.mock import patch
from pathlib import Path

from .make_mocks import (
    make_students_csv,
    make_parents_csv,
    random_class
)

with open(
    Path(Path(__file__).parent, 'random_names.csv'),
    'r'
) as csvf:
    rd = csv.reader(csvf)
    names = [r for r in rd]

# smaller sample for fast test
@ patch('teacherhelper.tests.make_mocks.names', names[:100])
class TestMakeMocks(unittest.TestCase):

    def test_make_students_csv(self):
        data = make_students_csv()

        assert data  # no empty list
        for row in data:
            # check that student names come from list of random names
            assert ' '.join(row[:2]) in [' '.join(i) for i in names]

            # check that none of the teachers are also students
            teacher_last, teacher_first = row[3].split(', ')
            assert ' '.join((teacher_first, teacher_last)) not in [
                ' '.join(i) for i in names
            ]

        # check that homerooms are all in the same grade
        groups_by_teacher = {}
        for row in data:
            groups_by_teacher.setdefault(row[3], set())
            groups_by_teacher[row[3]].add(row[2])
        for set_ in groups_by_teacher.values():
            assert len(set_) == 1


    def test_make_parents_csv(self):
        students = make_students_csv()
        parents = make_parents_csv(students)

        assert parents  # no empty list
        for row in parents:
            # every row has 13 items
            assert len(row) == 13
            # row is all strings
            for i in row:
                assert isinstance(i, str)
            # unpack parent name variables
            parent_first_nm, parent_last_nm, *_ = row
            parent_name = ' '.join((
                parent_first_nm,
                parent_last_nm
            ))
            # no parent shares a name with a student
            assert parent_name not in [' '.join(n) for n in names]
            # no parent shares a name with a teacher
            teachers = set([' '.join(r[3].split(', ')[::-1]) for r in students])
            assert parent_name not in teachers
        # two parents for each student
        assert len(parents) / 2 == len(students)

    def test_random_class(self):
        """
        Test that students from a random class are actually all in the same
        class and grade level.
        """
        students = make_students_csv()
        class_ = random_class(students)
        for strow in class_:
            assert strow[2] == class_[0][2]
            assert strow[3] == class_[0][3]



"""
Script to generate mock data.
"""

import csv
from pathlib import Path
from itertools import cycle
from typing import List
import random


with open(Path(Path(__file__).parent, 'random_names.csv'), 'r') as csvf:
    rd = csv.reader(csvf)
    names = [r for r in rd]


def make_students_csv() -> list:
    teachers = [
        ', '.join((n1[0], n2[1])) for n1, n2 in zip(names[:30], names[1:31])
    ]
    teacher_grade = {
        teacher : f'{random.randint(4, 6)}th Grade'
        for teacher in teachers
    }
    rows = []
    for teacher, name in zip(cycle(teachers), names):
        first, last = name
        # Each row is:
        # first, last, grade level, homeroom teacher, email, birthday
        rows.append((
            first,
            last,
            teacher_grade[teacher],
            teacher,
            str(hash(first + last))[1:] + '@empacad.org',
            f'{random.randint(1, 13)}/{random.randint(1,28)}/{random.randint(2005, 2010)}'
        ))
    return rows

def make_parents_csv(students_csv_data: list):
    """
    Need to take the student data to make random parents for those students.

    Each row has:
    - Guardian First Name
    - Guardian Last Name
    - Student First Name
    - Student Last Name
    - Primary Contact
    - Guardian Email Address 1
    - Guardian Mobile Phone
    - Guardian Phone
    - Guardian Work Phone
    - Comments
    - Allow Contact
    - Student Resides With
    - Relation to Student
    """
    parent_names = [
        (first[0], last[1]) for first, last in zip(cycle(names[50:]), names)
    ] * 2
    rows = []
    for parent_name, student_row in zip(
        parent_names,
        cycle(students_csv_data[5:])
    ):
        # fix any accidental duplicates
        while ' '.join(parent_name) in [' '.join(n) for n in names]:
            parent_name = (
                names[random.randint(0, len(names) - 1)][0],
                parent_name[1]
            )
        rows.append([
            *parent_name,                               # guardian full name
            student_row[0],                             # student first name
            student_row[1],                             # student last name
            'Y',                                        # is primary contact
            f'{str(hash(parent_name))[1:]}@gmail.com',  # email
            str(random.randint(9730000000, 9739999999)),       # mobile phone
            *[''] * 2,                                  # home & work phone
            str(hash(student_row)),                     # comments
            ['Y', 'N'][random.randint(0, 1)],                  # allow contact
            ['Y', 'N'][random.randint(0, 1)],                  # student resides with
            [
                'Mother',
                'Father',
                'Other',
                'Grandmother',
                'Grandfather'
            ][random.randint(0, 4)]                            # relation to student
        ])
    return rows

def random_class(students_csv_data: list) -> List[str]:
    teacher = random.sample(students_csv_data, 1)[0][3]
    return [s for s in students_csv_data if s[3] == teacher]





import os

from .helper import Helper

REQUIRED_VARS = [
    'EMAIL_USERNAME',
    'EMAIL_PASSWORD',
]

for var in REQUIRED_VARS:
    if not os.getenv(var):
        print(
            f'Environment variable {var} is missing, which is necessary for '
            'certain functions within this module.'
        )



# TODO: refactor as dataclasses

class EntityException(Exception): ...
class ParentGuardianError(EntityException): ...


class Group:
    """
    Group class is used to manage groups for extracurricular activities, field
    trips, and other purposes.
    """

    def __init__(self, name, grade_level, students):
        self.name = name
        self.grade_level = grade_level
        self.students = students


class Homeroom:

    def __init__(self, teacher, grade_level, students):
        """
        Ensure that string constants for csv headers of id, student names, and
        (if applicable) student emails are correct.
        """
        self.teacher = teacher
        self.grade_level = grade_level
        self.students = students


class Student:
    def __init__(self, context):
        context.setdefault('groups', [])
        context.setdefault('guardians', [])
        self.first_name = context.get('first_name')
        self.last_name = context.get('last_name')
        self.student_id = context.get('student_id')
        self.homeroom = context.get('homeroom')
        self.grade_level = context.get('grade_level')
        self.groups = context.get('groups')
        self.email = context.get('email')
        self.guardians = context.get('guardians')
        self.name = self.first_name + ' ' + self.last_name
        """
        primary_contact is an instance of ParentGuardian, which is assigned
        during the parsing of parent / guardian data in the new_school_year
        function within OnCourseMixin
        """
        self.primary_contact = None

    def __str__(self, verbose=False):
        """
        Assemble nice printout of student information, and student's guardian
        information. Also, offer verbose option, which ties into /shell.py
        """
        data = [
            '--',
            self.name,
            self.grade_level,
            self.homeroom,
            self.email,
            '--'
            '\nGuardians:'
        ]
        student_data = [str(i) for i in data if i]  # filter out NoneTypes
        if not self.guardians:
            return '\n'.join(data)
        # indent
        gu_data = ['\n\t'.join(self.primary_contact.__str__().split('\n'))]
        cutoff = 3 if not verbose else len(self.guardians) - 1
        for gu in self.guardians[1:cutoff]:
            gu_data.append('\n\t'.join(gu.__str__().split('\n')))
        return '\n'.join(student_data + gu_data)


class ParentGuardian:
    def __init__(self, context: dict, verbose=False):

        self.student                    = context.get('student')
        self.first_name                 = context.get('first_name')
        self.last_name                  = context.get('last_name')
        self.home_phone                 = context.get('home_phone')
        self.mobile_phone               = context.get('mobile_phone')
        self.work_phone                 = context.get('work_phone')
        self.email                      = context.get('email')
        self.relationship_to_student    = context.get('relationship_to_student')
        self.primary_contact            = context.get('primary_contact')
        self.allow_contact              = context.get('allow_contact')
        self.student_resides_with       = context.get('student_resides_with')

        # full name
        if self.first_name and self.last_name:
            self.name = self.first_name + ' ' + self.last_name
        else:
            raise ParentGuardianError(
                'First and last name were not provided as context to '
                'ParentGuardian class. This means that self.name cannot\n'
                'be constructed. Please at least pass a first and last name '
                'for every guardian into this class.'
            )

        # warn about missing attributes
        for k, v in self.__dict__.items():
            if not v:
                if verbose:
                    print(
                        f'WARNING: Guardian\t{self.name}\thas no value for\t{k}'
                    )

        # type checking
        if not isinstance(self.student, Student):
            raise ParentGuardianError(
                'Student was a string, not a Student object.'
            )
        for k, v in self.__dict__.items():
            if 'phone' in k and v and not isinstance(v, int):
                raise ParentGuardianError(
                    'Phone number should be of type int.'
                )

    def __str__(self):
        outs = [
            self.relationship_to_student,
            self.name,
            f'Contact allowed: {self.allow_contact}',
            f'Is primary contact: {self.primary_contact}',
            f'Mobile Phone: {self.mobile_phone}',
            f'Email Address: {self.email}',
            '\n',
        ]
        return '\n'.join([str(i) for i in outs if i])



from typing import Union, List, NamedTuple
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from pathlib import Path
import smtplib
import ssl
import os
import logging

from bs4 import BeautifulSoup

logger = logging.getLogger(__name__)

class Message(NamedTuple):
    html: str
    plain_text: str

class Email:
    def __init__(self, username=None, password=None):
        self.username:str = username if username else ''
        self.password:str = password if password else ''
        # use my default env variables for username and password if none were
        # passed
        if not self.username and (un := os.getenv('EMAIL_USERNAME')):
            self.username = un
        if not self.password and (pw := os.getenv('EMAIL_PASSWORD')):
            self.password = pw
        self.template_dir = Path(Path(__file__).parent, 'html_email_templates')
        self.connection = None

    def __enter__(self):
        context = ssl.create_default_context()
        self.connection = smtplib.SMTP_SSL(
            'smtp.gmail.com',
            port=465,
            context=context,
        )
        self.connection.login(
            self.username,
            self.password,
        )
        return self

    def __exit__(self, *_):
        if self.connection:
            self.connection.close()

    def send(self, *, to, subject, message: Union[Message, List[str]], cc=None, html=False):
        """
        If message is a list, its strings will be wrapped in <p> tags before
        being inserted into the email.

        If message is a tuple, it will be presumed to be 
        """
        if not self.connection:
            raise Exception(
                'Connection must be established in __enter__. Use this class '
                'as a context manager.'
            )
        # init ssl
        me = 'jdevries@empacad.org'
        msg = MIMEMultipart('alternative')
        msg['Subject'] = subject
        msg['From'] = me
        msg['To'] = to
        if cc:
            msg['Cc'] = cc
        if isinstance(message, list):
            html, plain_text = self.make_message(message)
        elif isinstance(message, Message):
            html, plain_text = message
        else:
            raise TypeError(
                f'Invalid message type: {type(message)}. Message must be of '
                'type list or teacherhelper.email.Message'
            )
        part1 = MIMEText(plain_text, 'plain')
        part2 = MIMEText(html, 'html')
        msg.attach(part1)
        msg.attach(part2)
        self.connection.sendmail(me, to, msg.as_string())

    def make_message(self, message: Union[list, str], *, template: Path=None):
        """
        Make a message with an html tempalte. A list will be interpreted
        as a list of paragraphs to be inserted while a string is expected
        to be html text ready to be fully inserted.
        """
        if not template:
            template = Path(self.template_dir, 'default.html')
        # load template
        with open(template, 'r') as htmlf:
            html_template = htmlf.read()

        if isinstance(message, list):
            # generate html paragraphs from list of strings
            html = []
            for i in message:
                if i:
                    html.append(
                        "<p style=\"color: black; font-family: Helvetica, Arial, Sans-Serif;\">"
                          f"{i}"
                        "</p>"
                    )
                else:
                    html.append('<br />')
            html = '\n'.join(html)
            message = '\n\n'.join(message)
        elif isinstance(message, str):
            # presume string should be directly inserted; already valid html
            html = message
            message = '\n'.join(BeautifulSoup(message, features='lxml').findAll(text=True))
        else:
            raise TypeError(
                f'Invalid message type: {type(message)}. Message must be of '
                'type list or str'
            )
        return (
            html_template.replace('{{ email_content }}', html),
            message
        )



import os
from pathlib import Path
import logging
import dbm
import shelve
from typing import Union
from datetime import datetime

from fuzzywuzzy import process

from .entities import Student
from .oncourse_mixin import OnCourseMixin


logger = logging.getLogger(__name__)


class HelperError(Exception): ...
class CacheError(HelperError): ...


class Helper(OnCourseMixin):
    """
    Driver for the entire module! See README.md test
    """

    DATA_DIR = (os.getenv('HELPER_DATA')
                or Path(Path(__file__).parents[1], 'private'))

    def __init__(self, homerooms: dict, students: dict, groups: dict):
        self.homerooms = homerooms
        self.students = students
        self.groups = groups
        self.cache_dir = os.path.join(__file__, 'cache')

    def write_cache(self):
        with shelve.open(os.path.join(self.DATA_DIR, 'cache'), 'c') as db:
            db['data'] = self
            db['date'] = datetime.now()

    def find_nearest_match(self, student_name: str, threshold=90, **_) -> Union[Student, None]:
        """
        Returns a student object. If auto_yes=True, it will presume that the
        best matching student is correct. Optionally, set a levenshtien distance
        threshold below which students will not be included.
        """
        if not isinstance(student_name, str):
            raise Exception("Student name must be a string")

        # direct match
        if st := self.students.get(student_name.title()):
            logger.debug(f'Exact match for {st.name}')
            return st

        # get nearest match
        if result := process.extractOne(
                student_name,
                self.students.keys()
        ):
            closest_name, confidence = result[0], result[1]
            if confidence >= threshold:
                return self.students[closest_name]

    def find_parent(self, parent_name):
        # create dict of guardians & primary_contacts
        all_guardians = {}
        primary_contacts = {}
        for st in self.students.values():
            try:
                primary = st.primary_contact
                primary_contacts.setdefault(primary.name, primary)
            except AttributeError:
                pass
            for g in st.guardians:
                all_guardians.setdefault(g.name, g)

        # prefer match amongst primary contacts
        primary_match = process.extractOne(
            parent_name,
            [g.name for g in primary_contacts.values()]
        )
        if primary_match and primary_match[1] > 85:
            if mo :=  primary_contacts.get(primary_match[0]):
                return mo.student

        # search all parents and guardians otherwise
        name_match = process.extractOne(
            parent_name,
            [g.name for g in all_guardians.values()]
        )
        if name_match and (mo := all_guardians.get(name_match[0])):
            return mo.student

    def exhaustive_search(self, name, subgroup=None, threshold=70):
        """
        Split the name into words and check if each word is a unique first or
        last name within the subgroup. If calling this function frequently
        on repeated datasets, try to cache the results because this function
        is slow.
        """
        # try matching normally first with default threshold of 90
        st = self.find_nearest_match(name, auto_yes=True)
        if st:
            logger.debug(f'Basic search for {name} returned {st.name}')
            return st
        logger.debug(f'Basic search for {name} returned None')
        if (st := self.search_within_subgroup(name,
                                              subgroup,
                                              threshold=threshold)):
            return st
        logger.debug(f'Full name subgroup search for {name} returned None')

        if not ' ' in name:
            name_parts = [name]
        else:
            name_parts = ' '.split(name)

        logger.debug(f'Searching through words {name_parts}')

        # search each name part as each role
        for np in name_parts:
            for role in ('last_name', 'first_name'):
                if (st := self.search_within_subgroup(np,
                                                      subgroup,
                                                      name_part=role,
                                                      threshold=threshold)):
                    logger.debug(f'{name} matched with {st.name} by {role}')
                    return st
        logger.debug(f'Exhaustive search for {name} failed')

    def search_within_subgroup(self, name, subgroup: list, threshold=90, name_part='name'):
        """
        subgroup should be a list of Student objects to search within. threshold
        is the confidence threshold. name_part is the part of the name that
        "name" is declared to be. Can be 'first_name', 'last_name', or just
        'name' (full name). At a higher level, this allows more or less
        exhaustive searching by calling this function multiple times.
        """
        logger.debug(f'Checking {name} as a {name_part}')
        name_match = process.extract(
            name,
            [getattr(s, name_part) for s in subgroup],
            limit=2
        )

        # Check whether name is unique within subgroup.
        if name_match[1][0] == name_match[0][0]:
            logger.debug(
                f'Cannot proceed with {name}. More than one '
                f'student in the subgroup has the first '
                f'name {name_match[0][0]}.'
            )

        # check for confidence
        if name_match[0][1] < threshold:
            logger.debug(
                f'{name_match[0][0]} rejected against {name} because the '
                f'confidence of {name_match[0][1]} was below the threshold of '
                + str(threshold)
            )
            return  # does not pass threshold

        name = name_match[0][0]

        if name == 'name':  # shortcut for full name matches
            if (st := self.students[name]):
                logger.debug('Hit on full name match shortcut')
                return st

        # we know the name is unique in the subgroup, so find the corresponding
        # Student object in the subgroup and return it.
        # TODO: refactor
        for st in self.students.values():

            if not st in subgroup:
                continue

            # break when we find the right student
            if name in st.name:
                break

        logger.debug(
            f'SUBGROUP MATCH {name} matches with {st.name} within subgroup.'
        )

        return st

    @ classmethod
    def read_cache(cls, check_date=True):
        """
        This static method returns a class because I like to break the rules.
        there's a reason for the rules; this garbage doesn't work
        """
        with shelve.open(os.path.join(cls.DATA_DIR, 'cache'), 'r') as db:
            cls = db['data']
            date = db['date']
        if check_date and (datetime.now().month in range(9, 12) and date.month in range(1, 7)):
            raise Exception(
                'It appears that the cache is from last school year. Please\n'
                'provide new data, re-instantiate, and re-write cache.'
            )
        return cls

    @ staticmethod
    def cache_exists():
        try:
            sh = shelve.open(os.path.join(Helper.DATA_DIR, 'cache'), 'r')
            sh.close()
            return True
        except dbm.error:  # type: ignore
            return False




from difflib import ndiff


class IterCsvError(Exception): ...


class IterCsv:
    def __init__(self, acceptable_headers, rows, strip_data=True):
        """
        Generator returns context, and row. Context is a dict in which
        the key is one of the acceptable headers, and the value is the
        index at which that header field can be found in each row.
        """
        self.current_row = 0
        self.rows = rows
        self.context = {}
        self._acceptable_headers = acceptable_headers
        self._strip_data = strip_data
        self._assign_context()

    def _assign_context(self):
        """Context is a mapping of header labels to row indices, so that the
        row items can later be fetched by name."""
        for i, raw_header in enumerate(self.rows[0]):
            raw_header = raw_header.lower()
            for clean_header in self._acceptable_headers:
                if clean_header == raw_header:
                    # create mapping and check for duplicate match
                    before = len(self.context)
                    self.context[clean_header] = i
                    after = len(self.context)
                    if before == after:
                        raise IterCsvError(
                            'There are two or more headers with the value '
                            f'{raw_header}. Edit the csv to differentiate '
                            'between these columns to continue.'
                        )
        if not (a := list(self.context.keys())) == (b := self._acceptable_headers):
            diff = '\n\t'.join(ndiff(a, b))
            msg = ('A match was not found for all headers:\n'
                   f'DIFF:\n\t{diff}')
            raise IterCsvError(msg)
        self._validate_context(self.context)

    def fetch(self, name: str):
        """Fetch an item from a row by name during iteration."""
        if (index := self.context.get(name)) is None:
            raise IterCsvError(f'{name} does not exist in csv context')
        value = self.rows[self.current_row][index]
        if self._strip_data:
            return value.strip()
        return value

    def __iter__(self):
        return self

    def __next__(self):
        self.current_row += 1
        if self.current_row < len(self.rows) - 1:
            return self.fetch
        raise StopIteration

    @ staticmethod
    def _validate_context(context):
        """Because this field is misspelled in OnCourse....."""
        is_parent_spreadsheet = 'guardian first name' in context
        is_header_misspelled = context.get('student resides with') is None

        if is_parent_spreadsheet and is_header_misspelled:
            raise IterCsvError(
                'Remember, "student resides with"  is misspelled in '
                'OnCourse. Fix it in the CSV you downloaded.'
            )
